package eu.kanade.tachiyomi.extension.en.hentai2read

import eu.kanade.tachiyomi.network.GET
import eu.kanade.tachiyomi.network.POST
import eu.kanade.tachiyomi.source.model.*
import eu.kanade.tachiyomi.source.online.ParsedHttpSource
import eu.kanade.tachiyomi.util.asJsoup
import okhttp3.FormBody
import okhttp3.OkHttpClient
import okhttp3.Request
import okhttp3.Response
import org.jsoup.nodes.Document
import org.jsoup.nodes.Element
import java.util.Calendar
import java.util.regex.Pattern

class Hentai2Read : ParsedHttpSource() {

    override val name = "Hentai2Read"

    override val baseUrl = "https://hentai2read.com"

    override val lang = "en"

    override val supportsLatest = true

    override val client: OkHttpClient = network.cloudflareClient

    companion object {
        val imageBaseUrl = "https://static.hentaicdn.com/hentai"

        val pagesUrlPattern by lazy {
            Pattern.compile("""'images' : \[\"(.*?)[,]?\"\]""")
        }

        val chapterDatePattern by lazy {
            Pattern.compile("""about (\d+\s+\w+\s+ago)""")
        }

        lateinit var base64String: String
    }

    override fun popularMangaSelector() = "div.img-container div.img-overlay a"

    override fun latestUpdatesSelector() = "ul.nav-users li.ribbon"

    override fun popularMangaRequest(page: Int)
        = GET("$baseUrl/hentai-list/all/any/all/most-popular/$page/", headers)

    override fun latestUpdatesRequest(page: Int)
        = GET("$baseUrl/latest/$page/", headers)

    override fun popularMangaFromElement(element: Element): SManga {
        val manga = SManga.create()
        manga.setUrlWithoutDomain(element.attr("href"))
        element.select("h2.mangaPopover").let {
            manga.title = it.attr("data-title").trim().substringBefore(" [").trim()
        }
        return manga
    }

    override fun latestUpdatesFromElement(element: Element): SManga {
        val manga = SManga.create()
        element.select("a.mangaPopover").let {
            manga.setUrlWithoutDomain(it.attr("href"))
            manga.title = it.attr("data-title").trim().substringBefore(" [").trim()
        }
        return manga
    }

    override fun popularMangaNextPageSelector() = "a#js-linkNext"

    override fun latestUpdatesNextPageSelector() = popularMangaNextPageSelector()

    override fun searchMangaRequest(page: Int, query: String, filters: FilterList): Request {
        val form = FormBody.Builder().apply {
            add("cmd_wpm_wgt_mng_sch_sbm", "Search")
            add("txt_wpm_wgt_mng_sch_nme", "")
            add("cmd_wpm_pag_mng_sch_sbm", "")
            add("txt_wpm_pag_mng_sch_nme", query)

            for (filter in if (filters.isEmpty()) getFilterList() else filters) {
                when (filter) {
                    is MangaNameSelect -> add("cbo_wpm_pag_mng_sch_nme", filter.state.toString())
                    is ArtistName -> add("txt_wpm_pag_mng_sch_ats", filter.state)
                    is ArtistNameSelect -> add("cbo_wpm_pag_mng_sch_ats", filter.state.toString())
                    is CharacterName -> add("txt_wpm_pag_mng_sch_chr", filter.state)
                    is CharacterNameSelect -> add("cbo_wpm_pag_mng_sch_chr", filter.state.toString())
                    is ReleaseYear -> add("txt_wpm_pag_mng_sch_rls_yer", filter.state)
                    is ReleaseYearSelect -> add("cbo_wpm_pag_mng_sch_rls_yer", filter.state.toString())
                    is Status -> add("rad_wpm_pag_mng_sch_sts", filter.state.toString())
                    is TagSearchMode -> add("rad_wpm_pag_mng_sch_tag_mde", arrayOf("and", "or")[filter.state])
                    is TagList -> filter.state.forEach { tag -> 
                        when (tag.state) {
                            Filter.TriState.STATE_INCLUDE -> add("chk_wpm_pag_mng_sch_mng_tag_inc[]", tag.id.toString())
                            Filter.TriState.STATE_EXCLUDE -> add("chk_wpm_pag_mng_sch_mng_tag_exc[]", tag.id.toString())
                        }
                    }
                }
            }
        }

        var searchUrl = "$baseUrl/hentai-list/advanced-search"
        if (page > 1) { searchUrl += "/${base64String}" }

        return POST("${searchUrl}/all/name-az/$page", headers, form.build())
    }

    override fun searchMangaParse(response: Response): MangasPage {
        val document = response.asJsoup()
        var hasNextPage = false

        val mangas = document.select(searchMangaSelector()).map { element ->
            searchMangaFromElement(element)
        }

        val nextPage = document.select(searchMangaNextPageSelector())
        nextPage.first()?.let {
            hasNextPage = true
            base64String = it.attr("href").substringAfter("/advanced-search/").substringBefore("/")
        }

        return MangasPage(mangas, hasNextPage)
    }

    override fun searchMangaSelector() = popularMangaSelector()

    override fun searchMangaFromElement(element: Element): SManga {
        return popularMangaFromElement(element)
    }

    override fun searchMangaNextPageSelector() = popularMangaNextPageSelector()

    override fun mangaDetailsParse(document: Document): SManga {
        val infoElement = document.select("ul.list-simple-mini").first()

        val manga = SManga.create()
        manga.author = infoElement.select("li:contains(Author) > a")?.text()
        manga.artist = infoElement.select("li:contains(Artist) > a")?.text()
        manga.genre = infoElement.select("li:contains(Category) > a, li:contains(Content) > a").map {
            it.text()
        }.joinToString(", ")
        manga.description = infoElement.select("li:contains(Storyline) > p")?.text()
        manga.status = infoElement.select("li:contains(Status) > a")?.text().orEmpty().let {parseStatus(it)}
        manga.thumbnail_url = document.select("a#js-linkNext > img")?.attr("src")
        return manga
    }

    private fun parseStatus(status: String) = when {
        status.contains("Ongoing") -> SManga.ONGOING
        status.contains("Completed") -> SManga.COMPLETED
        else -> SManga.UNKNOWN
    }

    override fun chapterListSelector() = "ul.nav-chapters > li > div.media > a"

    override fun chapterFromElement(element: Element): SChapter {
        val chapter = SChapter.create()
        chapter.setUrlWithoutDomain(element.attr("href"))
        chapter.name = element.ownText().trim()
        chapter.date_upload = element.select("div > small").text()?.let {
            val matcher = chapterDatePattern.matcher(it)
            if (matcher.find()) {
                parseChapterDate(matcher.group(1))
            } else {
                0L
            }
        } ?: 0L
        return chapter
    }

    private fun parseChapterDate(date: String): Long {
        val dateWords = date.split(" ")
        if (dateWords.size == 3) {
            val timeAgo = Integer.parseInt(dateWords[0])
            return Calendar.getInstance().apply {
                when (dateWords[1]) {
                    "minute", "minutes" -> Calendar.MINUTE
                    "hour", "hours" -> Calendar.HOUR
                    "day", "days" -> Calendar.DAY_OF_YEAR
                    "week", "weeks" -> Calendar.WEEK_OF_YEAR
                    "month", "months" -> Calendar.MONTH
                    "year", "years" -> Calendar.YEAR
                    else -> null
                }?.let {
                    add(it, -timeAgo)
                }
            }.timeInMillis
        }
        return 0L
    }

    override fun pageListRequest(chapter: SChapter) = POST(baseUrl + chapter.url, headers)

    override fun pageListParse(response: Response): List<Page> {
        val pages = mutableListOf<Page>()
        val m = pagesUrlPattern.matcher(response.body().string())
        var i = 0
        while (m.find()) {
            m.group(1).split(",").forEach {
                pages.add(Page(i++, "", imageBaseUrl + it.trim('"').replace("""\/""", "/")))
            }
        }
        return pages
    }

    override fun pageListParse(document: Document): List<Page> {
        throw Exception("Not used")
    }

    override fun imageUrlRequest(page: Page) = GET(page.url)

    override fun imageUrlParse(document: Document) = ""

    private class MangaNameSelect : Filter.Select<String>("Manga Name", arrayOf("Contains", "Starts With", "Ends With"))
    private class ArtistName : Filter.Text("Artist")
    private class ArtistNameSelect : Filter.Select<String>("Artist Name", arrayOf("Contains", "Starts With", "Ends With"))
    private class CharacterName : Filter.Text("Character")
    private class CharacterNameSelect : Filter.Select<String>("Character Name", arrayOf("Contains", "Starts With", "Ends With"))
    private class ReleaseYear : Filter.Text("Release Year")
    private class ReleaseYearSelect : Filter.Select<String>("Release Year", arrayOf("In", "Before", "After"))
    private class Status : Filter.Select<String>("Status", arrayOf("Any", "Completed", "Ongoing"))
    private class TagSearchMode : Filter.Select<String>("Tag Search Mode", arrayOf("AND", "OR"))
    private class Tag(name: String, val id: Int) : Filter.TriState(name)
    private class TagList(title: String, tags: List<Tag>) : Filter.Group<Tag>(title, tags)

    override fun getFilterList() = FilterList(
            MangaNameSelect(),
            Filter.Separator(),
            ArtistName(),
            ArtistNameSelect(),
            Filter.Separator(),
            CharacterName(),
            CharacterNameSelect(),
            Filter.Separator(),
            ReleaseYear(),
            ReleaseYearSelect(),
            Filter.Separator(),
            Status(),
            Filter.Separator(),
            TagSearchMode(),
            Filter.Separator(),
            TagList("Categories", getCategoryList()),
            Filter.Separator(),
            TagList("Tags", getTagList()),
            Filter.Separator(),
            TagList("Doujins", getDoujinList())
    )

    // Categories : 27
    // $("div#tab-category > div:has(a.block)").map((i, el) => `Tag("${$(el).select("a").first().text().trim()}", ${$(el).find("input").first().attr("value")})`).get().sort().join(",\n")
    // on https://hentai2read.com/hentai-search/"
    private fun getCategoryList() = listOf(
        Tag("Adult", 34),
        Tag("Anal", 7),
        Tag("Big Breasts", 20),
        Tag("Comedy", 43),
        Tag("Compilation", 46),
        Tag("Doujinshi", 42),
        Tag("Ecchi", 40),
        Tag("Futanari", 14),
        Tag("Gender Bender", 26),
        Tag("Harem", 31),
        Tag("Incest", 15),
        Tag("Lactation", 16),
        Tag("Licensed", 50),
        Tag("Lolicon", 17),
        Tag("Masturbation", 18),
        Tag("Oneshot", 33),
        Tag("Rape", 23),
        Tag("Robotics", 30),
        Tag("Romance", 41),
        Tag("School Life", 48),
        Tag("Serialized", 32),
        Tag("Shotacon", 44),
        Tag("Tentacles", 24),
        Tag("Tragedy", 49),
        Tag("Un-censored", 47),
        Tag("Yaoi", 27),
        Tag("Yuri", 28)
    )

    // Tags : 355
    // $("div#tab-tag > div:has(a.block)").map((i, el) => `Tag("${$(el).select("a").first().text().trim()}", ${$(el).find("input").first().attr("value")})`).get().sort().join(",\n")
    // on https://hentai2read.com/hentai-search/"
    private fun getTagList() = listOf(
        Tag("Abortion", 529),
        Tag("Absent Parents", 1423),
        Tag("Abusive Lover", 878),
        Tag("Abusive", 1587),
        Tag("Adapted to H-Anime", 416),
        Tag("Addiction", 1438),
        Tag("Adopted Sister", 1634),
        Tag("Adoption", 522),
        Tag("Adoptive Siblings", 1451),
        Tag("Adultery", 661),
        Tag("Affair", 1459),
        Tag("Aggressive Lover", 1599),
        Tag("Ahegao", 1702),
        Tag("Airheads", 1807),
        Tag("All-Girls School", 347),
        Tag("Alternative Ending", 666),
        Tag("Anal Play", 1365),
        Tag("Analingus (Rimjob)", 2298),
        Tag("Angels", 1093),
        Tag("Animal Girls", 1904),
        Tag("Animal Transformation", 710),
        Tag("Anthology", 589),
        Tag("Anthropomorphism", 913),
        Tag("Apron", 975),
        Tag("Armpit Licking", 2360),
        Tag("Armpit Sex", 1843),
        Tag("Arranged Marriage", 846),
        Tag("Artificial Intelligence", 1719),
        Tag("Assjob", 2190),
        Tag("Aunt-Nephew Relationship", 327),
        Tag("Aunts", 1661),
        Tag("Authority Figures", 1821),
        Tag("BBW", 1867),
        Tag("BDSM", 831),
        Tag("BSDM", 1263),
        Tag("Bad Grammar", 2188),
        Tag("Bathroom Intercourse", 350),
        Tag("Beach", 403),
        Tag("Best Friends", 1227),
        Tag("Bestiality", 372),
        Tag("Betrayal", 610),
        Tag("Big Ass", 1591),
        Tag("Bikini", 1514),
        Tag("Bishoujo", 645),
        Tag("Bittersweet Ending", 1439),
        Tag("Blackmail", 391),
        Tag("Blind Characters", 1875),
        Tag("Blindfold", 1177),
        Tag("Bloomers", 1444),
        Tag("Blow job", 952),
        Tag("Body Modification", 1760),
        Tag("Body Swap", 444),
        Tag("Body Writing", 2183),
        Tag("Bondage", 317),
        Tag("Borderline H", 395),
        Tag("Brainwash", 1321),
        Tag("Breast Expansion", 2191),
        Tag("Brides", 1351),
        Tag("Brother Complex", 574),
        Tag("Brother and Sister", 1305),
        Tag("Brother-in-law", 1782),
        Tag("Bukkake", 551),
        Tag("Bullying", 600),
        Tag("Bunny Girls", 1226),
        Tag("CGs", 1512),
        Tag("Cat Ears", 1633),
        Tag("Cat Girls", 957),
        Tag("Caught in the Act", 339),
        Tag("Censored", 1196),
        Tag("Cervix Penetration", 2184),
        Tag("Character Who Bullies the One They Love", 549),
        Tag("Cheating", 351),
        Tag("Cheerleaders", 1339),
        Tag("Child Abuse", 1852),
        Tag("Child Born From Incest", 575),
        Tag("Child Prostitute", 1663),
        Tag("Childhood Friends", 309),
        Tag("Childhood Love", 310),
        Tag("Chinese Dress", 2344),
        Tag("Chubby", 1819),
        Tag("Club President", 1586),
        Tag("Clumsy Character", 1808),
        Tag("Co-workers", 1739),
        Tag("Collection of Inter-Linked Stories", 415),
        Tag("Collection of Short Stories/Oneshots", 352),
        Tag("Confession", 834),
        Tag("Corruption", 755),
        Tag("Cosplay", 379),
        Tag("Cousins", 1340),
        Tag("Cow Girls", 1371),
        Tag("Creampie", 1037),
        Tag("Crossdressing", 343),
        Tag("Cunnilingus", 1754),
        Tag("Dark Skin", 1277),
        Tag("Debt-Motivated Prostitution", 1364),
        Tag("Debts", 1154),
        Tag("Deception", 516),
        Tag("Deep Throat", 1436),
        Tag("Defloration", 1246),
        Tag("Delinquents", 680),
        Tag("Demon Girls", 1453),
        Tag("Demon Hunters", 1666),
        Tag("Demons", 1152),
        Tag("Doctor-Patient Relationship", 900),
        Tag("Dog Girls", 353),
        Tag("Double Penetration", 427),
        Tag("Drugs", 446),
        Tag("Drunk Intercourse", 438),
        Tag("Drunk", 864),
        Tag("Elf-Elves", 1370),
        Tag("Emotionless Sex", 2312),
        Tag("Enema Play", 1740),
        Tag("Enemies Become Lovers", 803),
        Tag("Ero-Guro", 302),
        Tag("Exhibitionism", 404),
        Tag("Facesitting", 1831),
        Tag("Fairy-Fairies", 1368),
        Tag("Family Love", 1616),
        Tag("Family Secrets", 1623),
        Tag("Father and Daughter", 1021),
        Tag("Father-in-Law", 1922),
        Tag("Female Dominance", 320),
        Tag("Fetish", 1030),
        Tag("Fight Between Lovers", 1745),
        Tag("Fingering", 1389),
        Tag("First Love", 670),
        Tag("Fisting", 428),
        Tag("Foot job", 1058),
        Tag("Forced Marriage", 1617),
        Tag("Forced Sex", 1320),
        Tag("Forced into a Relationship", 1746),
        Tag("Foursome", 330),
        Tag("Fox Girls", 1268),
        Tag("Friends Become Lovers", 1769),
        Tag("Full Color", 468),
        Tag("Furry", 2315),
        Tag("Futa on Female", 2309),
        Tag("Futa on Male", 1850),
        Tag("Gang Rape", 429),
        Tag("Gangbang", 462),
        Tag("Ganguro", 696),
        Tag("Giantess", 2278),
        Tag("Girls Only", 1960),
        Tag("Glasses", 465),
        Tag("God-Human Relationship", 720),
        Tag("Goddess", 1309),
        Tag("Group Intercourse", 311),
        Tag("Gyaru", 2185),
        Tag("Hairy Armpit", 2359),
        Tag("Hand Job", 534),
        Tag("Happy Sex", 491),
        Tag("Hardcore", 1397),
        Tag("Harem-seeking Male Lead", 1487),
        Tag("Hot Springs", 1744),
        Tag("Housewife-Housewives", 555),
        Tag("Human Pet", 1853),
        Tag("Human Toilets", 1724),
        Tag("Human-Nonhuman Relationship", 459),
        Tag("Humiliation", 552),
        Tag("Hypnotism", 1732),
        Tag("Idols", 1211),
        Tag("Impregnation", 1358),
        Tag("Incest as a Subplot", 753),
        Tag("Infidelity", 423),
        Tag("Inflation", 2186),
        Tag("Inverted Nipples", 431),
        Tag("Jealous Lover", 441),
        Tag("Jealousy", 331),
        Tag("Kidnapping", 408),
        Tag("Kimono", 1283),
        Tag("Korean Comic", 2158),
        Tag("Kunoichi (Ninja Girls)", 1759),
        Tag("Large Dicks", 990),
        Tag("Leotard", 2187),
        Tag("Lingerie", 332),
        Tag("Little Sisters", 1580),
        Tag("Live-in Lover", 1771),
        Tag("Love At First Sight", 1326),
        Tag("Love Rivals", 1490),
        Tag("Love Triangles", 1167),
        Tag("MILFs", 354),
        Tag("Magic", 539),
        Tag("Magical Girls", 565),
        Tag("Maids", 912),
        Tag("Male Dominance", 681),
        Tag("Mangaka", 659),
        Tag("Masochists", 662),
        Tag("Master-Pet Relationship", 880),
        Tag("Master-Servant Relationship", 460),
        Tag("Master-Slave Relationship", 389),
        Tag("Mermaids", 1714),
        Tag("Mind Break", 303),
        Tag("Mind Control", 304),
        Tag("Molesters", 1114),
        Tag("Monster Girls", 1409),
        Tag("Monster Sex", 1243),
        Tag("Monsters", 1175),
        Tag("Mother Complex", 342),
        Tag("Mother and Daughter", 784),
        Tag("Mother and Son", 393),
        Tag("Mother-in-Law", 1716),
        Tag("Multiple Penetration", 1826),
        Tag("Neighbors", 1741),
        Tag("Netorare", 370),
        Tag("Netori", 2243),
        Tag("Newlyweds", 451),
        Tag("Nipple Intercourse", 1717),
        Tag("Nipple Play", 944),
        Tag("Non-Human Pregnancy", 461),
        Tag("Nuns", 1185),
        Tag("Nurses", 1046),
        Tag("Obsessive Love", 605),
        Tag("Office Ladies", 1738),
        Tag("Old Man", 2331),
        Tag("Older Brother", 1366),
        Tag("Older Female Young Boy", 1648),
        Tag("Older Female Younger Male", 355),
        Tag("Older Male Younger Female", 1229),
        Tag("Older Sister", 1241),
        Tag("Outdoor Intercourse", 504),
        Tag("Paizuri", 360),
        Tag("Pantyhose", 1275),
        Tag("Partial Censorship", 2303),
        Tag("Partially Colored", 665),
        Tag("Pegging", 1736),
        Tag("Personality Change", 576),
        Tag("Perverted Boss", 387),
        Tag("Perverted Characters", 300),
        Tag("Perverted Teachers", 366),
        Tag("Perverts", 1526),
        Tag("Pets", 1328),
        Tag("Piercings", 1465),
        Tag("Plain Girls", 1657),
        Tag("Plastic Surgery", 688),
        Tag("Polygamy", 312),
        Tag("Poor Characters", 1317),
        Tag("Poor Grammar", 2282),
        Tag("Porn Industry", 1440),
        Tag("Porn Stars", 367),
        Tag("Porn with Plot", 530),
        Tag("Possessed", 1656),
        Tag("Possession", 677),
        Tag("Possessive Lover", 1602),
        Tag("Pregnancy", 305),
        Tag("Pretend Rape", 1543),
        Tag("Priestesses", 1662),
        Tag("Priests", 1140),
        Tag("Princes", 1160),
        Tag("Princesses", 1161),
        Tag("Prisoners", 618),
        Tag("Proactive Protagonist", 1608),
        Tag("Producers", 1528),
        Tag("Prostitution", 394),
        Tag("Public Intercourse", 507),
        Tag("Public Nudity", 406),
        Tag("Punishment Sex", 787),
        Tag("Punishment", 1418),
        Tag("Queens", 1434),
        Tag("Rabbit Girls", 470),
        Tag("Reverse Harem", 567),
        Tag("Reverse Rape", 376),
        Tag("Rewrite", 2281),
        Tag("Rich Boy", 726),
        Tag("Rich Family", 727),
        Tag("Rich Girl", 368),
        Tag("Rushed Ending/Axed", 505),
        Tag("Sadist", 663),
        Tag("Sadomasochism", 499),
        Tag("Scat", 432),
        Tag("School Girls", 995),
        Tag("School Intercourse", 313),
        Tag("School Nurse-Student Relationship", 478),
        Tag("Secret Crush", 1491),
        Tag("Secret Relationship", 399),
        Tag("Seduction", 1556),
        Tag("Senpai-Kouhai Relationship", 443),
        Tag("Sex Addicts", 301),
        Tag("Sex Friends Become Lovers", 400),
        Tag("Sex Friends", 629),
        Tag("Sex Industry", 450),
        Tag("Sex Slaves", 1273),
        Tag("Sex Toys", 837),
        Tag("Sexual Abuse", 695),
        Tag("Sexual Assault", 476),
        Tag("Sexual Frustration", 1611),
        Tag("Shemale", 2325),
        Tag("Shy Characters", 652),
        Tag("Sibling Love", 655),
        Tag("Sister Complex", 440),
        Tag("Sister and Brother", 1403),
        Tag("Sisters", 1330),
        Tag("Sketchy Art Style", 811),
        Tag("Sleep Intercourse", 583),
        Tag("Sluts", 1377),
        Tag("Small Breasts", 434),
        Tag("Son Complex", 909),
        Tag("Spirits", 1387),
        Tag("Star-Crossed Lover/s", 1346),
        Tag("Step-Daughter", 531),
        Tag("Step-Father", 532),
        Tag("Step-Father/Step-Daughter Relationship", 533),
        Tag("Step-Mother", 770),
        Tag("Step-Mother/Step-Son Relationship", 334),
        Tag("Step-Sibling Love", 371),
        Tag("Step-Siblings", 521),
        Tag("Step-Son", 833),
        Tag("Stockings", 1327),
        Tag("Student Council", 641),
        Tag("Student-Tutor Relationship", 315),
        Tag("Succubus", 378),
        Tag("Sudden Appearance", 609),
        Tag("Sudden Confession", 425),
        Tag("Swimsuit/s", 1179),
        Tag("Tanned", 1479),
        Tag("Teacher-Student Relationship", 1126),
        Tag("Teacher-Teacher Relationship", 369),
        Tag("Teachers", 1388),
        Tag("Threesome (MFF)", 1686),
        Tag("Threesome (MMF)", 1688),
        Tag("Threesome (Other)", 335),
        Tag("Time Stop", 2300),
        Tag("Tomboy", 650),
        Tag("Torture", 2189),
        Tag("Transgender", 518),
        Tag("Trap", 344),
        Tag("Tribadism", 553),
        Tag("Tsundere", 653),
        Tag("Tutors", 591),
        Tag("Twincest", 578),
        Tag("Twins", 336),
        Tag("Uncle and Niece", 1923),
        Tag("Unlucky Character/s", 414),
        Tag("Unrequited Love", 1284),
        Tag("Unusual Pupils", 2313),
        Tag("Urethral Intercourse", 1157),
        Tag("Urethral Play", 817),
        Tag("Urination", 435),
        Tag("Vampires", 1136),
        Tag("Virgins", 1137),
        Tag("Virtual Reality", 588),
        Tag("Vore", 1841),
        Tag("Voyeurism", 361),
        Tag("Waitresses", 1334),
        Tag("Werewolf", 877),
        Tag("Widow", 544),
        Tag("Wife Corruption", 613),
        Tag("Wife Depravity", 545),
        Tag("Wife-Wives", 1385),
        Tag("Witches", 1485),
        Tag("Wolf Girls", 1412),
        Tag("X-Ray", 1071),
        Tag("Yandere", 873),
        Tag("Youkai", 1029),
        Tag("Young Master", 500),
        Tag("Yuri as a Subplot", 2342)
    )

    // Doujins : 868
    // $("div#tab-doujin > div:has(a.block)").map((i, el) => `Tag("${$(el).select("a").first().text().trim()}", ${$(el).find("input").first().attr("value")})`).get().sort().join(",\n")
    // on https://hentai2read.com/hentai-search/"
    private fun getDoujinList() = listOf(
        Tag("3-gatsu no Lion", 2350),
        Tag("3x3 Eyes", 1118),
        Tag("7th Dragon", 1401),
        Tag("81diver", 1880),
        Tag("A Channel", 2168),
        Tag("Accel World", 1584),
        Tag("Ace Attorney", 1990),
        Tag("Action Heroine: Cheer Fruits", 2339),
        Tag("Agent Aika", 2141),
        Tag("Ah! My Goddess", 1027),
        Tag("Ai yori Aoshi", 1522),
        Tag("Aikatsu!", 1892),
        Tag("Air", 2163),
        Tag("Aiura", 1885),
        Tag("Aiyoku no Eustia", 1983),
        Tag("Akame ga Kill!", 2042),
        Tag("Akaneiro Ni Somaru Saka", 2150),
        Tag("Akatsuki no Yona", 2295),
        Tag("Aldnoah Zero", 2111),
        Tag("Alice in Wonderland", 345),
        Tag("Alice in the Country of Hearts", 1951),
        Tag("Amaama To Inazuma", 2216),
        Tag("Amaeta Gari", 1876),
        Tag("Amagami", 937),
        Tag("Amagi Brilliant Park", 1989),
        Tag("Amanchu!", 2276),
        Tag("Amano Megumi Ha Sukidarake", 2326),
        Tag("And Yet The Town Moves", 1916),
        Tag("Ane Doki", 1699),
        Tag("Angel Beats!", 1080),
        Tag("Animal Crossing", 2364),
        Tag("Anne Happy", 2317),
        Tag("Ano Hana", 1090),
        Tag("Ano Hi Mita Hana no Namae o Bokutachi wa Mada Shiranai", 1109),
        Tag("Ano Natsu de Matteru", 1470),
        Tag("Another", 1776),
        Tag("Ansatsu Kyoushitsu", 2064),
        Tag("Ao no Exorcist", 1562),
        Tag("Aoi Hana", 1548),
        Tag("Aoi Shiro", 2044),
        Tag("Aquarion Evol", 1508),
        Tag("Ar Tonelico 2", 1578),
        Tag("Arakawa Under the Bridge", 1467),
        Tag("Arc The Lad", 2250),
        Tag("Arcana Heart", 1217),
        Tag("Aria", 1110),
        Tag("Arpeggio of Blue Steel", 1895),
        Tag("Arslan Senki", 2269),
        Tag("Asatte no Houkou", 2266),
        Tag("Ashitaba-san Chi no Muko Kurashi", 2299),
        Tag("Asobi ni Ikuyo", 956),
        Tag("Astarotte no Omocha!", 1705),
        Tag("Astro Fighter Sunred", 2338),
        Tag("Asu no Yoichi!", 967),
        Tag("Atelier Series", 2023),
        Tag("Axis Powers Hetalia", 2256),
        Tag("Azazel-san.", 1773),
        Tag("Azumanga Daioh", 1775),
        Tag("Azur Lane", 2368),
        Tag("Baccano!", 1980),
        Tag("Baka to Test to Shoukanjuu", 940),
        Tag("Bakemonogatari", 1024),
        Tag("Bakuman", 985),
        Tag("Bakunyu", 1244),
        Tag("Bakuon", 2214),
        Tag("Bamboo Blade", 1654),
        Tag("Bang Dream", 2333),
        Tag("Banner of the Stars", 2199),
        Tag("Basquash!", 1087),
        Tag("Bastard!!", 1593),
        Tag("Battle Girl High School", 2221),
        Tag("Battle Spirits", 1045),
        Tag("Bayonetta", 1011),
        Tag("Beat Angel Escalayer", 1701),
        Tag("Beat Blades Haruka", 2287),
        Tag("Beatmania", 1993),
        Tag("Beelzebub", 2102),
        Tag("Ben-To", 1785),
        Tag("Berserk", 1735),
        Tag("Bible Black", 2207),
        Tag("Big Hero 6", 2247),
        Tag("Bikini Warriors", 2198),
        Tag("Bishoujo Kamen Poitrine", 2162),
        Tag("Black Bullet", 1999),
        Tag("Black Butler", 2363),
        Tag("Black Cat", 1103),
        Tag("Black Desert Online", 2321),
        Tag("Black Lagoon", 1249),
        Tag("Black Rock Shooter", 1802),
        Tag("Blast of Tempest", 2083),
        Tag("Blazblue", 1523),
        Tag("Bleach", 933),
        Tag("Blue Exorcist", 2027),
        Tag("Boku wa Tomodachi ga Sukunai", 954),
        Tag("Bokutachi wa Benkyou ga Dekinai", 2366),
        Tag("Boruto", 2351),
        Tag("Brave Beats", 2206),
        Tag("Brave Witches", 2320),
        Tag("Bravely Default", 1902),
        Tag("Brothers Conflict", 1947),
        Tag("Brynhildr in the Darkness", 1944),
        Tag("Btooom", 2026),
        Tag("Buddy Complex", 1982),
        Tag("Busou Renkin", 1324),
        Tag("Busou Shinki", 1801),
        Tag("Busou Shoujo Machiavellianism", 2334),
        Tag("CANAAN", 2037),
        Tag("Campione!", 1706),
        Tag("Canvas", 2318),
        Tag("Captain Earth", 2146),
        Tag("Card Captor Sakura", 1919),
        Tag("Cardfight!! Vanguard", 2020),
        Tag("Catherine", 1692),
        Tag("Charlotte", 2108),
        Tag("Chihayafuru", 2072),
        Tag("Chobits", 1899),
        Tag("Chousoku Henkei Gyrozetter", 1767),
        Tag("Chrome Shelled Regios", 1787),
        Tag("Chrono Trigger", 1811),
        Tag("Chu Berozu", 1399),
        Tag("Chuunibyou demo Koi ga Shitai!", 1766),
        Tag("City Hunter", 1959),
        Tag("Clannad", 955),
        Tag("Claymore", 2030),
        Tag("Code Geass", 1009),
        Tag("Collar x Malice", 2343),
        Tag("Corpse Party", 2124),
        Tag("Cowboy Bebop", 1958),
        Tag("Crayon Shin-chan", 2011),
        Tag("Cutey Honey", 2314),
        Tag("Cyborg 009", 1410),
        Tag("D-Fragments", 1242),
        Tag("Da Capo", 1723),
        Tag("Dagashi Kashi", 2053),
        Tag("Daily Life with a Monster Girl", 1781),
        Tag("Dakara Boku wa H ga Dekinai", 1881),
        Tag("Danball Senki", 1954),
        Tag("Dance With Devils", 2209),
        Tag("Danganronpa", 1896),
        Tag("Danshi Koukousei No Nichijou", 2130),
        Tag("Daphne in the Brilliant Blue", 1332),
        Tag("Darker Than Black", 964),
        Tag("Darkstalkers", 1857),
        Tag("Date A Live", 1794),
        Tag("Dead Dead Demons Dededededestruction", 2077),
        Tag("Dead or Alive", 1040),
        Tag("DearS", 2039),
        Tag("Death Note", 1188),
        Tag("Demi-chan wa Kataritai", 2268),
        Tag("Demons Souls", 1635),
        Tag("Dennou Coil", 1018),
        Tag("Denpa Onna to Seishun Otoko", 1091),
        Tag("Detective Conan", 2051),
        Tag("Devil Survivor 2", 1884),
        Tag("Devilman", 2046),
        Tag("Di Gi Charat", 2093),
        Tag("Diabolik Lovers", 2117),
        Tag("Digimon Savers", 2218),
        Tag("Digimon Xros Wars", 1102),
        Tag("Dirty Pair no Daibouken", 901),
        Tag("Dirty Pair", 1260),
        Tag("Discipline", 2358),
        Tag("Disgaea 2", 1877),
        Tag("Doctor Strange", 2340),
        Tag("Dog Days", 1051),
        Tag("Doki Doki Majo Shinpan", 1266),
        Tag("Dokidoki! Precure", 1854),
        Tag("Dokkiri Doctor", 2076),
        Tag("Doraemon", 2147),
        Tag("Dorei To No Seikatsu", 2277),
        Tag("Dorohedoro", 1258),
        Tag("Double Sensei Life", 2240),
        Tag("Doutei", 1712),
        Tag("Dr. Grayman", 1197),
        Tag("Dragon Ball", 1122),
        Tag("Dragon Quest", 936),
        Tag("Dragonaut", 1059),
        Tag("Dragons Crown", 1865),
        Tag("Dream C Club", 939),
        Tag("Drifters", 2022),
        Tag("Dry Humping", 871),
        Tag("Dungeon Meshi", 2104),
        Tag("Dungeon ni Deai o Motomeru no wa Machigatteiru Darou ka", 2040),
        Tag("Durarara!!", 1337),
        Tag("Dynasty Warriors", 1610),
        Tag("Dystopia", 810),
        Tag("Elsword", 2200),
        Tag("Emma", 1012),
        Tag("Endless Frontier", 2238),
        Tag("Ensemble Stars!", 2246),
        Tag("Ero Manga Sensei", 2290),
        Tag("Etrian Odyssey", 1977),
        Tag("Euphoria", 2235),
        Tag("Eureka Seven", 337),
        Tag("Excel Saga", 2286),
        Tag("Eyeshield 21", 1207),
        Tag("Fairy Tail", 1113),
        Tag("Fallout", 2357),
        Tag("Fancy Lala", 1265),
        Tag("Fatal Frame", 2242),
        Tag("Fate Apocrypha", 2203),
        Tag("Fate Extella", 2329),
        Tag("Fate Grand Order", 2148),
        Tag("Fate Hollow Ataraxia", 915),
        Tag("Fate Kaleid Liner Prisma Illya", 1964),
        Tag("Fate-Extra CCC", 1790),
        Tag("Fate-Stay Night", 1042),
        Tag("Fate-Zero", 1638),
        Tag("Final Dragon Chronicle", 1421),
        Tag("Final Fantasy", 1201),
        Tag("Final Fight", 1125),
        Tag("Fire Emblem", 1205),
        Tag("Flip Flappers", 2289),
        Tag("Flower Knight Girl", 2252),
        Tag("Flying Witch", 2308),
        Tag("Free!", 2100),
        Tag("Freezing", 1847),
        Tag("Fresh Precure!", 993),
        Tag("Fruits Basket", 2119),
        Tag("Fujiyama-san wa Shishunki", 1966),
        Tag("Full Metal Daemon Muramasa", 1675),
        Tag("Full Metal Panic", 1119),
        Tag("Fullmetal Alchemist", 1107),
        Tag("Fun Fun Pharmacy", 2129),
        Tag("Fushigiboshi no Futagohime", 2071),
        Tag("Fushoku No Machi", 2161),
        Tag("Futari wa Precure Splash Star", 1698),
        Tag("Futari wa Precure", 1707),
        Tag("Futari wa Pretty Cure", 1903),
        Tag("Future Card Buddyfight", 2355),
        Tag("GJ Club", 1813),
        Tag("GJ-bu", 2180),
        Tag("Gabriel Dropout", 2270),
        Tag("Gad Guard", 2271),
        Tag("Gakkou Gurashi!", 2145),
        Tag("Gakusen Toshi Asterisk", 2156),
        Tag("Galaxy Angel", 1416),
        Tag("Galilei Donna", 2113),
        Tag("Gantz", 2249),
        Tag("Gatchaman Crowds", 1979),
        Tag("Gate - Jietai Kano Chi Nite Kaku Tatakaeri", 2114),
        Tag("Gegege No Kitarou", 2157),
        Tag("Gekkan Shoujo Nozaki-kun", 1974),
        Tag("Genmu Senki Leda", 2144),
        Tag("Genshiken Nidaime", 1890),
        Tag("Genshiken", 1245),
        Tag("Getsuyoubi no Tawawa", 2241),
        Tag("Getter Robo", 1883),
        Tag("Ghost In The Shell", 1092),
        Tag("Ghost Sweeper Mikami", 2239),
        Tag("Gintama", 1952),
        Tag("Girl Friend Beta", 2254),
        Tag("Girls Frontline", 2362),
        Tag("Girls and Panzer", 1798),
        Tag("Girl’s High", 1998),
        Tag("Go! Princess PreCure", 2054),
        Tag("Gochuumon wa Usagi desu ka?", 1968),
        Tag("God Eater", 961),
        Tag("Gosick", 2092),
        Tag("Granblue Fantasy", 2109),
        Tag("Grisaia no Kajitsu", 2041),
        Tag("Gugure! Kokkuri-san", 2179),
        Tag("Guilty Crown", 1789),
        Tag("Guilty Gear", 1254),
        Tag("Guity Crown", 1786),
        Tag("Gundam ZZ", 1957),
        Tag("Gundam", 996),
        Tag("Gunparade March", 2032),
        Tag("Gunslinger Girl", 2138),
        Tag("Gunsmith Cats", 2120),
        Tag("Guzuguzu Shitetara Sodacchau yo?", 1984),
        Tag("Hacka Doll", 2234),
        Tag("Hai to Gensou no Grimgal", 2228),
        Tag("Haikyu!!", 1996),
        Tag("Haiyore! Nyaruko-san", 1938),
        Tag("Hakuouki", 2294),
        Tag("Hanamaru Youchien", 1319),
        Tag("Hanasaku Iroha", 1026),
        Tag("Hanayamata", 2297),
        Tag("Happinesscharge Precure!", 1946),
        Tag("Harry Potter", 1291),
        Tag("Harukanaru Jikuu no Naka de", 2006),
        Tag("Harvest Moon", 2292),
        Tag("Hataraku Maou-sama!", 1839),
        Tag("Hatsukoi Limited", 969),
        Tag("Hayate no Gotoku", 1065),
        Tag("He is My Master", 1280),
        Tag("Heartcatch​ Precure​!", 1791),
        Tag("Hellsing", 2248),
        Tag("Hentai Ouji to Warawanai Neko", 1868),
        Tag("Heroic Age", 2346),
        Tag("Hibike! Euphonium", 2088),
        Tag("Hidamari Sketch", 1417),
        Tag("High School DxD", 1833),
        Tag("High Score Girl", 2352),
        Tag("Highschool of the Dead", 991),
        Tag("Higurashi no Naku Koro ni", 1096),
        Tag("Hikaru No Go", 2255),
        Tag("Himawari no Shoujo", 1192),
        Tag("Himouto! Umaru-chan", 2078),
        Tag("Historys Strongest Disciple Kenichi", 1041),
        Tag("Hitsugi no Chaika", 1976),
        Tag("Hokenshitsu no Shinigami", 1561),
        Tag("Hoshi no Samidare", 2125),
        Tag("Houkago Play", 999),
        Tag("Hundred", 2262),
        Tag("Hunter x Hunter", 1129),
        Tag("Hyakka Ryouran Samurai Girls", 1693),
        Tag("Hyouka", 1762),
        Tag("Hyperdimension Neptunia", 1879),
        Tag("I''s", 1281),
        Tag("IGPX", 2001),
        Tag("Ichiban Ushiro no Daimaou", 1335),
        Tag("Ichigo 100%", 1278),
        Tag("Ichigo Mashimaro", 2201),
        Tag("Ikki Tousen", 946),
        Tag("Imouto ga Iru!", 1727),
        Tag("Inazuma Eleven", 1912),
        Tag("Inda no Himekishi Janne", 1718),
        Tag("Infinite Ryvius", 2121),
        Tag("Infinite Stratos", 920),
        Tag("Inju Seisen Twin Angel", 1975),
        Tag("Inou-Battle wa Nichijou-kei no Naka de", 2224),
        Tag("Inu x Boku SS", 1950),
        Tag("Inuyasha", 1985),
        Tag("Irresponsible Captain Tylor", 2348),
        Tag("Isekai no Seikishi Monogatari", 1509),
        Tag("Ixion Saga DT", 2091),
        Tag("Jewelpet Sunshine", 1874),
        Tag("Jigoku Sensei Nube", 1765),
        Tag("Jinrui Wa Suitai Shimashita", 2123),
        Tag("Jitsu wa Watashi wa", 2310),
        Tag("JoJos Bizarre Adventures", 1534),
        Tag("Joker Game", 2274),
        Tag("Jormungand", 1770),
        Tag("Joshiraku", 1873),
        Tag("Journey to the West", 786),
        Tag("Jungle Guu", 2372),
        Tag("Juuza Engi", 2068),
        Tag("K-ON!", 945),
        Tag("Kaiten Mutenmaru", 2210),
        Tag("Kaitou Tenshi Twin Angel", 2126),
        Tag("Kakegurui", 2324),
        Tag("Kaleido Star", 1202),
        Tag("Kamen Rider W", 1463),
        Tag("Kami Nomi zo Shiru Sekai", 1069),
        Tag("Kami-sama no Inai Nichiyoubi", 2127),
        Tag("Kami-sama no Memochou", 1891),
        Tag("Kamichu", 2135),
        Tag("Kamikaze Kaitou Jeanne", 1788),
        Tag("Kampfer", 1187),
        Tag("Kannagi", 1184),
        Tag("Kannazuki No Miko", 2137),
        Tag("Kanokon", 1546),
        Tag("Kanon", 983),
        Tag("Kantai Collection", 1870),
        Tag("Kara no Kyoukai", 1900),
        Tag("Karakai Jouzu no Takagi-san", 2371),
        Tag("Kashimashi", 1757),
        Tag("Katanagatari", 1806),
        Tag("Katekyo Hitman Reborn", 2132),
        Tag("Kemono Ekaki no Kousoku 2", 1864),
        Tag("Kemono Friends", 2279),
        Tag("Kemono of Magic - Foxy Rena", 2283),
        Tag("Kenzen Robo Daimidaier", 2009),
        Tag("Keroro Gunsou", 1336),
        Tag("KiMiKiSS", 934),
        Tag("Kid Icarus", 1721),
        Tag("Kiddy Grade", 1869),
        Tag("Kill la Kill", 1893),
        Tag("Kimagure Orange Road", 1582),
        Tag("Kimi ga Aruji de Shitsuji ga Ore de", 1259),
        Tag("Kimi no Na wa", 2230),
        Tag("Kindaichi Shounen no Jikenbo", 2231),
        Tag("King of Fighters", 1124),
        Tag("Kiniro Mosaic", 1855),
        Tag("Kino No Tabi", 2347),
        Tag("Kirakira Precure a la Mode", 2341),
        Tag("Kiss x Sis", 1733),
        Tag("Kizuato", 1497),
        Tag("Kobayashi-san Chi no Maid Dragon", 2261),
        Tag("Kochikame", 1986),
        Tag("Kodomo no Jikan", 2143),
        Tag("Koe de Oshigoto!", 1780),
        Tag("Koe no Katachi", 2021),
        Tag("Koi to Senkyo to Chocolate", 1749),
        Tag("Koihime Musou", 1097),
        Tag("Kokoro Connect", 1848),
        Tag("Kono Bijutsubu ni wa Mondai ga Aru!", 2336),
        Tag("Kono Naka ni Hitori", 1726),
        Tag("Kono Subarashii Sekai Ni Syukufuku O", 2139),
        Tag("Kore wa Zombie desu ka?", 1844),
        Tag("Koukou Kyuuji Zawa-san", 1764),
        Tag("Koutetsujou No Kabaneri", 2169),
        Tag("Kuma Miko", 2155),
        Tag("Kure-nai", 2014),
        Tag("Kurogane no Liberals", 1671),
        Tag("Kurokami", 2134),
        Tag("Kuroko no Basuke", 1931),
        Tag("Kuromukuro", 2302),
        Tag("Kuttsukiboshi", 1820),
        Tag("Kyo Kara Maoh!", 2128),
        Tag("Kyonyuu Fantasy", 2305),
        Tag("Kyoukai Senjou no Horizon", 941),
        Tag("Kyoukai no Kanata", 2074),
        Tag("Kyuujou Lovers", 2223),
        Tag("La Blue Girl", 2353),
        Tag("Last Period", 2311),
        Tag("Le Fruit de la Grisaia", 2012),
        Tag("League of Legends", 1720),
        Tag("Leed Sha", 418),
        Tag("Legend Of Queen Opala", 2361),
        Tag("Legend of Mana", 1949),
        Tag("Legend of the Mystical Ninja", 1926),
        Tag("Linebarrels of Iron", 1672),
        Tag("Little Busters", 2107),
        Tag("Little Red Riding Hood", 2322),
        Tag("Log Horizon", 1914),
        Tag("Lotte no Omocha!", 1834),
        Tag("Love Hina", 1533),
        Tag("Love Lab", 1836),
        Tag("Love Live Sunshine", 2213),
        Tag("Love Live! - School Idol Project", 1917),
        Tag("Love Plus", 931),
        Tag("Lucky Star", 1000),
        Tag("Luminous Arc", 1310),
        Tag("Lupin III", 2116),
        Tag("Mabinogi", 1909),
        Tag("Macross Delta", 2244),
        Tag("Macross Frontier", 1098),
        Tag("Madan no Ou to Vanadis", 2090),
        Tag("Made In Abyss", 2328),
        Tag("Magi - Labyrinth of Magic", 1729),
        Tag("Magic Knight Rayearth", 2285),
        Tag("Magic the Gathering", 1816),
        Tag("Magical Girl Lyrical Nanoha", 1469),
        Tag("Magical Taluluto-kun", 1935),
        Tag("Maho Girls Precure", 2165),
        Tag("Mahou Sensei Negima", 1019),
        Tag("Mahou Shoujo Ai", 2316),
        Tag("Mahou Shoujo Ikusei Keikaku", 2253),
        Tag("Mahou Shoujo Lyrical Nanoha StrikerS", 1431),
        Tag("Mahou Shoujo Madoka Magica", 942),
        Tag("Mahou Shoujo Nante Mou Ii Desu Kara", 2275),
        Tag("Mahou Tsukai no Yoru", 1596),
        Tag("Mahouka Koukou no Rettousei", 1810),
        Tag("Mai-Hime", 1247),
        Tag("Maison Ikkoku", 2070),
        Tag("Maji de Watashi ni Koishinasai", 1673),
        Tag("Majin Tantei Nougami Neuro", 1583),
        Tag("Major", 2080),
        Tag("Makai Kishi Ingrid", 2115),
        Tag("Makai Senki Disgaea", 1898),
        Tag("Maken-Ki!", 1818),
        Tag("Mamono Musume Zukan", 2345),
        Tag("Manyuu Hikenchou", 2013),
        Tag("Maoyuu Maou Yuusha", 1768),
        Tag("Maria+Holic", 1005),
        Tag("Maria-sama ga Miteru", 926),
        Tag("Martian Successor Nadesico", 2245),
        Tag("Marvel Universe", 1927),
        Tag("Mashiroiro Symphony", 1563),
        Tag("Mass Effect", 2219),
        Tag("Mayo Chiki", 1088),
        Tag("Maze", 2217),
        Tag("Mazinger Z", 2225),
        Tag("Medaka Box", 1121),
        Tag("MegaMan", 2015),
        Tag("Melancholy of Haruhi Suzumiya", 1297),
        Tag("Melty Blood", 1419),
        Tag("Metal Slug", 2059),
        Tag("Metroid", 1973),
        Tag("Mikakunin de Shinkoukei", 1969),
        Tag("Minami-ke", 2075),
        Tag("Mirai Nikki", 1032),
        Tag("Mitsudomoe", 1924),
        Tag("Mobile Suit Gundam", 1948),
        Tag("Moetan", 2154),
        Tag("Mondaiji-tachi ga Isekai kara Kuru Sou Desu yo?", 2031),
        Tag("Monster Girl Quest", 2029),
        Tag("Monster Hunter", 921),
        Tag("Moshidora", 2195),
        Tag("Mouretsu Pirates", 1685),
        Tag("Moyashimon", 1965),
        Tag("Muramasa", 1039),
        Tag("Mushibugyo", 2065),
        Tag("Mushishi", 2018),
        Tag("Musou Orochi Z", 1797),
        Tag("Muv-Luv Alternative", 1550),
        Tag("Muv-Luv", 1547),
        Tag("My Hero Academia", 2057),
        Tag("My Little Pony", 2060),
        Tag("Myriad Colors Phantom World", 2226),
        Tag("Mysterious Girlfriend X", 1722),
        Tag("Nadia", 930),
        Tag("Nagi no Asukara", 1971),
        Tag("Namiuchigiwa no Muromi-san", 2081),
        Tag("Nanatsu no Taizai", 2033),
        Tag("Naruto", 914),
        Tag("Natsu no Kumo", 1379),
        Tag("Nausicaa", 1374),
        Tag("Nazo no Kanojo X", 1007),
        Tag("Negima", 1219),
        Tag("Nekomonogatari", 1866),
        Tag("Neon Genesis Evangelion", 938),
        Tag("Netoge no Yome wa Onnanoko ja Nai to Omotta?", 2170),
        Tag("New Game", 2181),
        Tag("Nichijou", 1887),
        Tag("NieR RepliCant", 2097),
        Tag("Ninja Gaiden Sigma", 1391),
        Tag("Ninja Slayer", 2229),
        Tag("Nisekoi", 1730),
        Tag("Nisemonogatari", 1779),
        Tag("No Game No Life", 1972),
        Tag("No More Heroes", 2003),
        Tag("No-Rin", 1908),
        Tag("Nodame Cantabile", 2319),
        Tag("Noein", 2233),
        Tag("Noir", 2367),
        Tag("Non Non Biyori", 1888),
        Tag("Noragami", 1945),
        Tag("Nurarihyon no Mago", 1400),
        Tag("Nyan Koi", 1086),
        Tag("Nyotaika! Paradise", 1889),
        Tag("OS-tan", 1067),
        Tag("Oboro Muramasa", 1933),
        Tag("Occult Academy", 1795),
        Tag("Occultic Nine", 2265),
        Tag("Oda Nobuna no Yabou", 1905),
        Tag("Odin Sphere", 1001),
        Tag("Ojamajo Doremi", 2008),
        Tag("Okujou no Yurirei-san", 2136),
        Tag("Okusan", 1731),
        Tag("One Off", 1851),
        Tag("One Piece", 917),
        Tag("One Punch Man", 1992),
        Tag("One Week Friends", 2002),
        Tag("Onegai My Melody", 1036),
        Tag("Onegai Teacher", 1918),
        Tag("Oni Imo", 1871),
        Tag("Onii-chan Dakedo Ai Sae Areba Kankeinai yo ne", 2236),
        Tag("Oniichan dakedo Ai Sae Areba Kankeinai yo ne", 1778),
        Tag("Ookami Kodomo no Ame to Yuki", 2103),
        Tag("Ookami to Koushinryou", 1055),
        Tag("Ookami-san to Shichinin no Nakama-tachi", 1115),
        Tag("Ookami-san", 976),
        Tag("Ookiku Furikabutte", 1674),
        Tag("Ooyasan Wa Shishunki", 2257),
        Tag("Ore Twintail ni Narimasu", 2034),
        Tag("Ore no Imouto ga Konna ni Kawaii Wake ga Nai", 1023),
        Tag("Ore no Kanojo to Osananajimi ga Shuraba Sugiru", 1872),
        Tag("Oshiete! Gyaruko-chan", 2152),
        Tag("Otome wa Boku ni Koishiteru", 1963),
        Tag("Otoyomegatari", 1937),
        Tag("Ouran High School Host Club", 2284),
        Tag("Outbreak Company", 1897),
        Tag("Outlaw Star", 1133),
        Tag("Overlord", 2176),
        Tag("Overwatch", 2196),
        Tag("Pandora Hearts", 1728),
        Tag("Pani Poni Dash", 1678),
        Tag("Panty and Stocking with Garterbelt", 1138),
        Tag("Panty and Stocking", 1450),
        Tag("Papa no Iu Koto o Kikinasai!", 1921),
        Tag("Persona 3", 1111),
        Tag("Persona 4", 977),
        Tag("Persona 5", 2215),
        Tag("Phantasy Star Online 2", 1882),
        Tag("Photo Kano", 1878),
        Tag("Pink Trash", 1516),
        Tag("Planetes", 1907),
        Tag("Plus Minus", 1048),
        Tag("Pokemon", 1402),
        Tag("Pretty Cure", 1035),
        Tag("Pretty Rhythm", 2264),
        Tag("PriPara", 1994),
        Tag("Prince of Tennis", 2258),
        Tag("Princess Lover!", 1827),
        Tag("Princess Resurrection", 1647),
        Tag("Prink Trash", 1755),
        Tag("Prison School", 2106),
        Tag("Prunus Girl", 2354),
        Tag("Pumpkin Scissors", 1407),
        Tag("Puzzle and Dragons", 1939),
        Tag("Queens Blade", 1120),
        Tag("Quiz Magic Academy", 1915),
        Tag("R.O.D the TV", 1859),
        Tag("RWBY", 2112),
        Tag("Rage of Bahamut", 2016),
        Tag("Ragnarok Online", 1150),
        Tag("Rail Wars", 2067),
        Tag("Rainy Days", 819),
        Tag("Rakudai Kishi no Eiyuutan", 2164),
        Tag("Rakuen Tsuihou", 2028),
        Tag("Rampo Kitan Game of Laplace", 2177),
        Tag("Rance Quest", 1970),
        Tag("Ranma 1-2", 986),
        Tag("Re:Zero kara Hajimeru Isekai Seikatsu", 2173),
        Tag("Read or Die", 1106),
        Tag("Real Drive", 2237),
        Tag("Reco Love", 2296),
        Tag("Record of Lodoss War", 1272),
        Tag("Renkin 3 Kyuu Magical? Pokaan", 1549),
        Tag("Resident Evil", 1089),
        Tag("Rewrite", 2251),
        Tag("Rinne no Lagrange", 1803),
        Tag("Rio: Rainbow Gate", 1708),
        Tag("Ro-Kyu-Bu! SS", 1925),
        Tag("Ro-Kyu-Bu", 2131),
        Tag("Robopon", 2167),
        Tag("Robot Ponkottsu", 1825),
        Tag("Rokka no Yuusha", 2079),
        Tag("Rokujouma no Shinryakusha!?", 2202),
        Tag("Rosario to Vampire", 1308),
        Tag("Rozen Maiden", 1828),
        Tag("Rumble Roses", 2192),
        Tag("Rune Factory", 2293),
        Tag("Rurouni Kenshin", 1415),
        Tag("Ryuugajou Nanana no Maizoukin", 2017),
        Tag("Sabagebu", 2098),
        Tag("Saber Marionette", 2004),
        Tag("Saenai Heroine no Sodatekata", 2043),
        Tag("Saijaku Muhai No Bahamut", 2304),
        Tag("Sailor Moon", 927),
        Tag("Saint Seiya", 2063),
        Tag("Saki Achiga-hen", 1934),
        Tag("Saki", 1066),
        Tag("Sakiko-san no Dansei Jijou", 1603),
        Tag("Sakura Taisen", 2160),
        Tag("Sakura Trick", 2069),
        Tag("Sakura War", 1683),
        Tag("Sakurasou no Pet na Kanojo", 1930),
        Tag("Samurai Flamenco", 2101),
        Tag("Samurai Spirits", 1128),
        Tag("Sangoku Rensenki", 1942),
        Tag("Sankarea", 2110),
        Tag("Saya no Uta", 1920),
        Tag("Sayonara Zetsubou Sensei", 1886),
        Tag("Scared Rider Xechs", 2212),
        Tag("School Rumble", 981),
        Tag("Seiken Densetsu 3", 1302),
        Tag("Seikon No Qwaser", 2118),
        Tag("Seitokai Yakuindomo", 1901),
        Tag("Sekai De Ichiban Tsuyoku Naritai", 2122),
        Tag("Sekai Seifuku", 1955),
        Tag("Sekirei", 1006),
        Tag("Selector Infected WIXOSS", 2047),
        Tag("Sengoku Basara", 1261),
        Tag("Sengoku Collection", 1962),
        Tag("Sengoku Musou", 1677),
        Tag("Sengoku Otome", 1911),
        Tag("Sengoku Rance", 1225),
        Tag("Senki Zesshou Symphogear", 2049),
        Tag("Sennen Sensou Aigis", 2227),
        Tag("Senran Kagura", 1443),
        Tag("Senyuu", 2263),
        Tag("Seraph of the End", 2232),
        Tag("Servant x Service", 1846),
        Tag("Seto no Hanayome", 1231),
        Tag("Shadow Hearts", 2025),
        Tag("Shakugan no Shana", 2096),
        Tag("Shaman King", 2048),
        Tag("Shantae", 2151),
        Tag("Sharin no Kuni", 1191),
        Tag("Shimoneta to Iu Gainen ga Sonzai Shinai Taikutsu na Sekai", 2133),
        Tag("Shin Koihime Musou", 1570),
        Tag("Shin Megami Tensei: Devil Survivor 2", 1835),
        Tag("Shingeki no Kyojin", 1832),
        Tag("Shining Blade", 1734),
        Tag("Shining Force", 2175),
        Tag("Shinken", 2335),
        Tag("Shinmai Fukei Kiruko-san", 1748),
        Tag("Shinmai Maou no Keiyakusha", 2084),
        Tag("Shinrabansho", 1842),
        Tag("Shinryaku! Ika Musume", 1953),
        Tag("Shinsekai Yori", 2220),
        Tag("Shirobako", 2045),
        Tag("Shirokuma Cafe", 2085),
        Tag("Shokugeki no Soma", 1956),
        Tag("Shota Eater", 1981),
        Tag("Shounen Maid", 2291),
        Tag("Show by Rock!!", 2095),
        Tag("Shugo Chara", 2019),
        Tag("Shuriken Sentai Ninninger", 2178),
        Tag("Shut Hell", 2073),
        Tag("Silent Hill", 2211),
        Tag("Simoun", 1936),
        Tag("Sket Dance", 2089),
        Tag("Skullgirls", 2159),
        Tag("Slayers", 1214),
        Tag("Smile Precure!", 1637),
        Tag("So-Ra-No-Wo-To", 2260),
        Tag("Solatorobo", 2052),
        Tag("Solty Rei", 1094),
        Tag("Sonic", 2055),
        Tag("Sora no Manimani", 2082),
        Tag("Sora no Method", 2094),
        Tag("Sora no Otoshimono", 1025),
        Tag("Sora o Kakeru Shoujo", 1193),
        Tag("Soredemo Machi wa Mawatteiru", 1174),
        Tag("Soreyuke Marin-chan", 1296),
        Tag("Soul Calibur", 992),
        Tag("Soul Eater", 1408),
        Tag("Sound Voltex", 2066),
        Tag("Sousei No Onmyouji", 2272),
        Tag("Space Battleship Yamato 2199", 1792),
        Tag("Space Dandy", 2062),
        Tag("Space Patrol Luluco", 2365),
        Tag("Spice and Wolf", 1411),
        Tag("Splatoon", 2099),
        Tag("Star Driver", 1172),
        Tag("Star Fox", 2327),
        Tag("Star Ocean 3", 1210),
        Tag("Star Ocean", 1655),
        Tag("Starry Sky", 2208),
        Tag("Steins;Gate", 1050),
        Tag("Stellvia Of The Universe", 1298),
        Tag("Stratos 4", 2036),
        Tag("Strawberry Panic", 1279),
        Tag("Street Fighters", 1081),
        Tag("Strike The Blood", 2222),
        Tag("Strike Witches", 1437),
        Tag("Suberu", 419),
        Tag("Suisei no Gargantia", 1814),
        Tag("Suite PreCure", 1020),
        Tag("Sukatto Golf Pangya", 1393),
        Tag("Summer Wars", 1131),
        Tag("Summon Night", 2323),
        Tag("Sumomomo Momomo", 1697),
        Tag("Sunohara-sou no Kanrinin-san", 2197),
        Tag("Super Black Jack", 1208),
        Tag("Super Danganronpa 2", 1988),
        Tag("Super Dimensional Fortress Macross", 2182),
        Tag("Super Mario", 1539),
        Tag("Super Powers", 824),
        Tag("Super Robot Wars", 1700),
        Tag("Super Sonico", 1774),
        Tag("Suzumiya Haruhi no Yuuutsu", 918),
        Tag("Sword Art Online", 1473),
        Tag("TERA Online", 1809),
        Tag("TERA The Exiled Realm of Arborea", 1894),
        Tag("Tactics Ogre: Wheel of Fate", 1747),
        Tag("Taimanin Asagi", 2035),
        Tag("Tales of Graces", 1753),
        Tag("Tales of Symphonia", 1845),
        Tag("Tales of Vesperia", 1743),
        Tag("Tales of Xillia", 1756),
        Tag("Tales of the Abyss", 928),
        Tag("Tamako Market", 1863),
        Tag("Tanken Driland", 2280),
        Tag("Tantei Opera Milky Holmes", 2174),
        Tag("Tari Tari", 1860),
        Tag("Tasogare Otome x Amnesia", 1750),
        Tag("Teen Titans", 1594),
        Tag("Tejina Senpai", 2205),
        Tag("Tekken", 1203),
        Tag("Tenchi Muyou", 1315),
        Tag("Tengen Toppa Gurren Lagann", 1112),
        Tag("Tenjo Tenge", 1108),
        Tag("Tenkai Knights", 2370),
        Tag("Tenkuu no Escaflowne", 2330),
        Tag("Terra Formars", 2153),
        Tag("The Idolm@sters", 948),
        Tag("The Legend of Heroes", 1987),
        Tag("The Legend of Zelda", 1793),
        Tag("The Lord of Elemental", 1689),
        Tag("The Ring", 2332),
        Tag("The Seven Deadly Sins", 2005),
        Tag("The World God Only Knows", 1123),
        Tag("They Are My Noble Masters", 1237),
        Tag("Tiger and Bunny", 1176),
        Tag("To Aru Kagaku no Railgun", 1557),
        Tag("To Aru Majutsu no Index", 960),
        Tag("To Heart 2", 987),
        Tag("To Heart", 984),
        Tag("To Love-Ru Darkness", 949),
        Tag("To Love-Ru", 1056),
        Tag("Tobaku Mokushiroku Kaiji", 1680),
        Tag("Toki o Kakeru Shoujo", 1913),
        Tag("Tokidensho Angel Eyes", 1480),
        Tag("Tokimeki Memorial 4", 1535),
        Tag("Tokimeki Memorial", 1812),
        Tag("Tokyo 7th Sisters", 2142),
        Tag("Tokyo Ghoul", 2056),
        Tag("Tokyo Ravens", 1961),
        Tag("Top wo Nerae 2!", 1372),
        Tag("Toradora!", 1095),
        Tag("Touhou", 916),
        Tag("Touken Ranbu", 2166),
        Tag("Trinity Seven", 2038),
        Tag("Tsukihime", 1446),
        Tag("Tsukumogami", 1077),
        Tag("Turn A Gundam", 1763),
        Tag("UN-GO", 1995),
        Tag("UQ Holder!", 1991),
        Tag("Uchi No Musume Ni Te O Dasuna", 2204),
        Tag("Uchuu Kaizoku Captain Harlock", 1189),
        Tag("Uchuu no Stellvia", 1289),
        Tag("Ultimate Girls", 2087),
        Tag("Umineko no Naku Koro ni", 1149),
        Tag("Umineko", 1910),
        Tag("Unbalance x Unbalance", 1405),
        Tag("Under Night In-Birth", 2301),
        Tag("Under the Moon", 2193),
        Tag("Undertale", 2356),
        Tag("Unlight", 1862),
        Tag("Upotte!!", 1592),
        Tag("Urusei Yatsura", 1466),
        Tag("Usagi Drop", 2105),
        Tag("Usavich", 1347),
        Tag("Uta no Prince-sama", 1929),
        Tag("Utawarerumono", 1062),
        Tag("Valkyria Chronicles II", 1643),
        Tag("Valkyria Chronicles", 925),
        Tag("Valkyrie Drive", 2259),
        Tag("Vampire Hunters", 1447),
        Tag("Vampire Princess Miyu", 2061),
        Tag("Vampire Savior", 1448),
        Tag("Vandread", 1462),
        Tag("Violated Heroine", 2172),
        Tag("Viper", 2007),
        Tag("Vividred Operation", 1838),
        Tag("Vocaloid", 1013),
        Tag("Voiceroid", 2306),
        Tag("Waiting in the Summer", 1906),
        Tag("Walkure Romanze", 2000),
        Tag("Warriors Orochi Z", 1684),
        Tag("Warship Girls", 2349),
        Tag("Watashi ga Motenai no wa Dou Kangaete mo Omaera ga Warui!", 1621),
        Tag("Wedding Peach", 2288),
        Tag("Wild Arms 2", 982),
        Tag("Wild Arms 3", 1804),
        Tag("Witch Blade", 997),
        Tag("Witch Craft Works", 924),
        Tag("Working", 1130),
        Tag("World Masterpiece Theater", 2050),
        Tag("World Trigger", 2149),
        Tag("Wrestle Angels", 1267),
        Tag("Xenoblade", 2307),
        Tag("Xenogears", 1777),
        Tag("Xenosaga", 1928),
        Tag("Yahari Ore no Seishun Rabukome wa Machigatte Iru.", 1796),
        Tag("Yakitate!! Japan", 1499),
        Tag("Yakushiji Ryouko no Kaiki Jikenbo", 2267),
        Tag("Yama no Susume", 1967),
        Tag("Yes! Precure 5", 1307),
        Tag("Yondemasu yo Azazeru-san", 1978),
        Tag("Yoru no Yatterman", 2171),
        Tag("Yosuga No Sora", 2140),
        Tag("Yotsuba!", 1234),
        Tag("Yotsubato!", 1943),
        Tag("You Are Under Arrest!", 2369),
        Tag("Youkoso Jitsuryoku Shijou Shugi no Kyoushitsu e", 2337),
        Tag("Yowamushi Pedal", 2086),
        Tag("Yu-Gi-Oh!", 1084),
        Tag("Yumekui Merry", 2024),
        Tag("Yuri On Ice", 2273),
        Tag("Yuru Yuri", 1861),
        Tag("Yuuki Yuuna wa Yuusha de Aru", 2058),
        Tag("Yuusha ni Narenakatta", 1932),
        Tag("Yuyushiki", 2010),
        Tag("Zegapain", 1530),
        Tag("Zero In", 1858),
        Tag("Zero no Tsukaima", 1148),
        Tag("Zettai Junpaku Mahou Shoujo", 2194),
        Tag("Zettai Karen Children", 1022),
        Tag("Zoids Genesis", 1052),
        Tag("Zone of the Enders", 1997)
    )
}
